import os
from multiprocessing import Pool
from tempfile import TemporaryDirectory

import pandas as pd
from tqdm.auto import tqdm

from era5epw.utils import (
    execute_download_request,
    make_cds_days_list,
    now_utc,
    unzip_and_load_netcdf_to_df,
)

url = "https://cds.climate.copernicus.eu/api"
datasets = [
    # new experimental dataset for timeseries data. Seems faster to download but not all variables are available.
    # first in the list so it is selected with higher priority.
    "reanalysis-era5-single-levels-timeseries",
    # new experimental dataset for land data. It's also faster to download but not all variables are available.
    "reanalysis-era5-land-timeseries",
    # the original dataset for single-level data. Contains all variables but is much slower to download.
    "reanalysis-era5-single-levels",
]
supported_vars_by_dataset = {
    "reanalysis-era5-single-levels-timeseries": [
        "2m_dewpoint_temperature",
        "2m_temperature",
        "total_precipitation",
        "10m_u_component_of_wind",
        "10m_v_component_of_wind",
        "surface_pressure",
    ],
    "reanalysis-era5-land-timeseries": [
        "soil_temperature_level_1",
    ],
    "reanalysis-era5-single-levels": ["*"],
}


def make_cds_request(
    ds: str | None,
    variables: [str],
    year: int,
    month: int | None,
    latitude: float,
    longitude: float,
    time_zone: int | None = None,
) -> list[dict[str, any]] | None:
    """Create a CDS request for the specified parameters.

    :param ds: The dataset to use, e.g., 'reanalysis-era5-single-levels-timeseries'.
    :param variables: List of variables to request.
    :param year: The year of the data.
    :param month: The month of the data. If None, all months will be requested.
    :param latitude: The latitude for the data point.
    :param longitude: The longitude for the data point.
    :param time_zone: Time zone offset from UTC. If provided, will adjust date range to
        fetch additional data needed for time zone conversion.
    :return: A dictionary representing the CDS request if the request is valid, otherwise
        None.
    """

    now = now_utc()

    # compute the start and end months.
    if year > now.year:
        return None
    if month is None:
        month_start = 1
        month_end = 12 if year < now.year else now.month
    else:
        if (now.year, now.month) < (year, month):
            return None
        month_start = month_end = month

    # dynamic dataset selection based on variables
    if ds is None:
        assert (
            len(variables) == 1
        ), "Dataset dynamic selection only supports single variable requests."
        for dataset in datasets:
            if variables[0] in supported_vars_by_dataset[dataset] or supported_vars_by_dataset[
                dataset
            ] == ["*"]:
                ds = dataset
                break

    if ds in [
        "reanalysis-era5-single-levels-timeseries",
        "reanalysis-era5-land-timeseries",
    ]:
        last_day_of_month_end = make_cds_days_list(year, month_end)[-1]
        start_date_str = f"{year}-{month_start:02d}-01"
        end_date_str = f"{year}-{month_end:02d}-{last_day_of_month_end}"

        main_request = {
            "dataset": ds,
            "data_format": "netcdf",
            "variable": variables,
            "date": [f"{start_date_str}/{end_date_str}"],
            "location": {"longitude": longitude, "latitude": latitude},
        }

        # Adjust date range based on time zone if provided.
        # We append an additional 1-day request to cover the shifted time zone
        if time_zone is not None:
            if time_zone >= 0 and month_start == 1:
                start_date_str = f"{year - 1}-12-31"
                end_date_str = f"{year - 1}-12-31"
            elif time_zone < 0 and month_end == 12:
                start_date_str = f"{year + 1}-01-01"
                end_date_str = f"{year + 1}-01-01"
            else:
                return [main_request]

            tz_request = main_request.copy()
            tz_request["date"] = [f"{start_date_str}/{end_date_str}"]
            return [main_request, tz_request]

        elif time_zone is None and month_start == 1:
            prev_day_request = main_request.copy()
            prev_day_request["date"] = [f"{year - 1}-12-31/{year - 1}-12-31"]
            return [prev_day_request, main_request]

        else:
            return [main_request]

    elif ds == "reanalysis-era5-single-levels":
        assert (
            month is not None
        ), "Month must be specified for 'reanalysis-era5-single-levels' dataset."

        main_request = {
            "dataset": ds,
            "product_type": "reanalysis",
            "data_format": "netcdf",
            "variable": variables,
            "year": [str(year)],
            "month": [f"{month:02d}"],
            "day": make_cds_days_list(year, month),
            "time": [f"{i:02d}:00" for i in range(24)],
            # North, West, South, East corner points of the area to extract.
            # Since 25th Feb 2026, point extraction isn't supported anymore on this dataset, we need to specify an area.
            # See:
            # https://forum.ecmwf.int/t/software-upgrade-for-data-extraction-of-a-geographical-area-from-selected-era5-and-seasonal-forecast-datasets/14583
            # https://confluence.ecmwf.int/display/CKB/Software+upgrade+for+geographical+area+extraction+from+data+on+regular+lat-lon+grids
            "area": [latitude + 0.1, longitude - 0.1, latitude - 0.1, longitude + 0.1],
        }

        # extend the request to cover time zone shifts
        if time_zone is not None:
            if time_zone >= 0 and month_start == 1:
                tz_request = main_request.copy()
                tz_request["year"] = [str(year - 1)]
                tz_request["month"] = ["12"]
                tz_request["day"] = ["31"]
                return [main_request, tz_request]
            elif time_zone < 0 and month_end == 12:
                tz_request = main_request.copy()
                tz_request["year"] = [str(year + 1)]
                tz_request["month"] = ["01"]
                tz_request["day"] = ["01"]
                return [main_request, tz_request]

        # add previous year's last day to fetch 00:00
        elif time_zone is None and month == 1:
            prev_day_request = main_request.copy()
            prev_day_request["year"] = [str(year - 1)]
            prev_day_request["month"] = ["12"]
            prev_day_request["day"] = ["31"]
            return [prev_day_request, main_request]

        return [main_request]

    else:
        raise ValueError(
            f"Unsupported dataset: {ds}. Supported datasets are 'reanalysis-era5-single-levels-timeseries' and 'reanalysis-era5-single-levels'."
        )


def make_intermediate_file_names(tmpdir: str, cds_requests: list[dict[str, any]]) -> list[str]:
    """Generate a list of temporary file names for storing intermediate results of CDS requests.

    :param tmpdir: Temporary directory where intermediate files will be stored.
    :param cds_requests: List of CDS requests that will be executed.
    :return: List of temporary file names.
    """
    # timeseries dataset requests have a 'date' field, while single-level requests have 'month'
    requested_years_months: set[tuple[int, int]] = set()
    for cds_request in cds_requests:
        if "date" in cds_request:
            date_range = cds_request["date"][0]
            start_date_str, end_date_str = date_range.split("/")
            start_year, start_month, _ = map(int, start_date_str.split("-"))
            _, end_month, _ = map(int, end_date_str.split("-"))
            for month in range(start_month, end_month + 1):
                requested_years_months.add((start_year, month))
        elif "month" in cds_request:
            month = int(cds_request["month"][0])
            year = int(cds_request["year"][0])
            requested_years_months.add((year, month))
        else:
            raise ValueError("CDS request must contain 'date' or 'month' field.")

    assert len(requested_years_months) > 0, "No months were requested in the CDS requests."
    years_months = sorted(list(requested_years_months))

    # extract the variable names from all CDS requests
    # not using a set to preserve the order of variables
    variables = []
    for cds_request in cds_requests:
        if "variable" in cds_request:
            for var in cds_request["variable"]:
                if var not in variables:
                    variables.append(var)
        else:
            raise ValueError("CDS request must contain 'variable' or 'variables' field.")

    intermediate_files = [
        os.path.join(tmpdir, f"era5_{year}_{month:02d}_{var}.nc")
        for year, month in years_months
        for var in variables
    ]

    return intermediate_files


def download_era5_data(
    variables: [str],
    year: int,
    latitude: float,
    longitude: float,
    dataset: str | None = datasets[0],
    parallel_exec_nb: int = 4,
    clean_up: bool = True,
    verbose: bool = False,
    time_zone: int | None = None,
) -> pd.DataFrame:
    """Download data from the Climate Data Store (CDS) for a specific variable and time and return
    as a DataFrame.

    :param variables: The variable to download (e.g., '2m_temperature').
    :param year: The year of the data. Full year will be downloaded.
    :param latitude: The latitude for the data point.
    :param longitude: The longitude for the data point.
    :param dataset: The dataset to use, e.g., 'reanalysis-era5-single-levels-timeseries'. If
        None, the first dataset supporting the variable will be selected.
    :param parallel_exec_nb: Number of parallel executions for downloading data. Default is
        12 (1 per month).
    :param clean_up: If True, remove individual month files after combining them into the
        target file.
    :param verbose: If True, enable verbose logging from CDS client.
    :param time_zone: Time zone offset from UTC. If provided, will adjust date range to
        fetch additional data needed for time zone conversion.
    :return: A DataFrame containing the downloaded data, combined on the 'time' dimension.
    """
    # split the request by month and variable, handle time zone adjustments
    cds_requests = [
        make_cds_request(
            ds=dataset,
            variables=[variable],
            year=year,
            month=month,
            latitude=latitude,
            longitude=longitude,
            time_zone=time_zone,
        )
        for month in range(1, 13)
        for variable in variables
    ]
    # flatten the list of lists and remove None entries
    cds_requests = [req for sublist in cds_requests for req in sublist if req is not None]

    if len(cds_requests) == 0:
        raise ValueError(
            f"No valid CDS requests could be created for year {year} and variables {variables}."
        )

    tqdm.write(
        f"Running a total of {len(cds_requests)} requests "
        f"with {parallel_exec_nb} parallel requests for {year}..."
    )

    # make temporary directory for intermediate files
    # then download each month in parallel
    with TemporaryDirectory(delete=clean_up) as tmpdir:
        intermediate_files = make_intermediate_file_names(tmpdir, cds_requests)

        assert len(cds_requests) == len(intermediate_files), (
            "The number of CDS requests and intermediate files must match."
            f" {len(cds_requests)} requests, {len(intermediate_files)} files."
        )

        # Create progress bar for ERA5 requests
        era5_progress = tqdm(
            total=len(cds_requests), desc="ERA5 requests", unit="request", position=1, leave=False
        )

        with Pool(parallel_exec_nb) as pool:
            # Use starmap_async to execute in parallel while maintaining progress tracking
            result = pool.starmap_async(
                execute_download_request,
                [
                    (
                        url,
                        cds_request["dataset"],
                        cds_request,
                        intermediary_file,
                        verbose,
                    )
                    for (cds_request, intermediary_file) in zip(cds_requests, intermediate_files)
                ],
            )

            # Poll for completion and update progress
            while not result.ready():
                # Count completed files to update progress
                completed = sum(1 for f in intermediate_files if os.path.exists(f))
                era5_progress.n = completed
                era5_progress.refresh()
                result.wait(timeout=1)  # Check every second

            # Final update
            era5_progress.n = len(cds_requests)
            era5_progress.refresh()
            _result = result.get()

        era5_progress.close()

        dfs = [
            unzip_and_load_netcdf_to_df(file_path, clean_up=clean_up)
            for file_path in intermediate_files
        ]

        # concatenate all DataFrames into a single DataFrame
        # 1. group by year and month, then concatenate along the variable axis
        dfs_by_year_month = {}
        for df in dfs:
            idx = (df.index[0].year, df.index[0].month)
            if idx not in dfs_by_year_month:
                dfs_by_year_month[idx] = []
            dfs_by_year_month[idx].append(df)

        dfs = [
            pd.concat(year_month_dfs, axis=1)
            for year_month_dfs in dfs_by_year_month.values()
            if len(year_month_dfs) > 0
        ]

        # 2. concatenate along the index (time) axis
        return pd.concat(dfs, axis=0, ignore_index=False).sort_index().drop_duplicates()


if __name__ == "__main__":
    df = download_era5_data(
        dataset=None,  # dynamic dataset selection based on variables
        variables=[
            "2m_temperature",
            "2m_dewpoint_temperature",
            "surface_pressure",
            "10m_u_component_of_wind",
            "10m_v_component_of_wind",
            "total_cloud_cover",
            "uv_visible_albedo_for_direct_radiation",
            "snow_depth",
            "total_precipitation",
            "soil_temperature_level_1",
        ],
        year=2024,
        latitude=49.4,
        longitude=0.1,
        clean_up=False,
        parallel_exec_nb=10,
        time_zone=1,
    )

    print(df.head(5))
    print(df.info())



import tempfile

import pandas as pd
import xarray as xr
from tqdm.auto import tqdm

from era5epw.utils import execute_download_request, now_utc

url = "https://ads.atmosphere.copernicus.eu/api"
dataset = "cams-solar-radiation-timeseries"


def make_cams_solar_radiation_request(
    longitude: float,
    latitude: float,
    year: int,
    sky_type: str = "observed_cloud",
    altitude: list[str] | None = None,
    time_step: str = "1hour",
    time_reference: str = "universal_time",
    time_zone: int | None = None,
) -> dict[str, any] | None:
    assert sky_type in [
        "clear",
        "observed_cloud",
    ], "Invalid sky type. Choose 'clear' or 'observed_cloud'."

    if altitude is None:
        altitude = ["0"]  # Default altitude if not specified

    now = now_utc()
    if year > now.year:
        return None  # Do not allow requests for future years

    # We start fetching at year - 1 even if time_zone is None as CAMS data starts at 01:00 UTC
    start_day = f"{year - 1}-12-31"
    end_day = f"{year}-12-31"
    # Adjust date range based on time zone if provided
    if time_zone is not None and time_zone < 0:
        end_day = f"{year + 1}-01-01"

    today = now.strftime("%Y-%m-%d")
    if end_day > today:
        end_day = today

    return {
        "sky_type": sky_type,
        "location": {"longitude": longitude, "latitude": latitude},
        "altitude": altitude,
        "date": [f"{start_day}/{end_day}"],
        "time_step": time_step,
        "time_reference": time_reference,
        "format": "netcdf",
    }


def download_cams_solar_radiation_data(
    longitude: float,
    latitude: float,
    year: int,
    sky_type: str = "observed_cloud",
    altitude: list[str] | None = None,
    time_step: str = "1hour",
    time_reference: str = "universal_time",
    clean_up: bool = True,
    time_zone: int | None = None,
) -> pd.DataFrame:
    """Download solar radiation data from the Copernicus Atmosphere Data Store (CAMS).

    :param longitude: Longitude of the location.
    :param latitude: Latitude of the location.
    :param year: Year for which to download the data.
    :param sky_type: Type of sky conditions, either "clear" or "observed_cloud".
    :param altitude: List of altitudes in meters, default is ["0"] if not specified.
    :param time_step: Time step for the data, default is "1hour".
    :param time_reference: Time reference for the data, default is "universal_time".
    :param clean_up: If True, remove the temporary file after processing.
    :param time_zone: Time zone offset from UTC. If provided, will adjust date range to
        fetch additional data needed for time zone conversion.
    """
    request = make_cams_solar_radiation_request(
        longitude=longitude,
        latitude=latitude,
        year=year,
        sky_type=sky_type,
        altitude=altitude,
        time_step=time_step,
        time_reference=time_reference,
        time_zone=time_zone,
    )

    if request is None:
        raise ValueError("Cannot download data for future years.")

    with tempfile.NamedTemporaryFile(dir="/tmp", suffix=".nc", delete=clean_up) as temp_file:
        # Create progress bar for CAMS request
        cams_progress = tqdm(total=1, desc="CAMS request", unit="request", position=1, leave=False)
        execute_download_request(
            url=url,
            dataset=dataset,
            cds_request=request,
            target_file=temp_file.name,
        )
        cams_progress.update(1)
        cams_progress.close()

        tqdm.write(f"Data downloaded to {temp_file.name}")

        ds = xr.open_dataset(temp_file.name)
        df = ds.to_dataframe()
        df.index = pd.to_datetime(df.index.get_level_values("time"))
        df = df.sort_index()

    return df


if __name__ == "__main__":
    df_2024 = download_cams_solar_radiation_data(
        longitude=2.69022,
        latitude=46.98709,
        year=2024,
        sky_type="observed_cloud",
        altitude=["0"],
        time_step="1hour",
        time_reference="universal_time",
    )
    print(df_2024.head(5))



import datetime
import glob
import logging
import os
import os.path
import random
import time
import zipfile
from base64 import b64encode
from calendar import monthrange
from pathlib import Path
from types import TracebackType

import cdsapi
import pandas as pd
import xarray as xr
from ecmwf.datastores import legacy_client

_api_key = None


class QuietEra5LegacyClientLoggingContext:
    """An override of the ecmwf.datastores.legacy_client.LoggingContext to set the logging level to
    ERROR."""

    def __init__(self, logger: logging.Logger, quiet: bool, debug: bool) -> None:
        self.old_level = logger.level
        if quiet:
            # only change compared to original LoggingContext (WARNING -> ERROR)
            logger.setLevel(logging.ERROR)
        else:
            logger.setLevel(logging.DEBUG if debug else logging.INFO)

        self.new_handlers = []
        if not logger.handlers:
            formatter = logging.Formatter("%(asctime)s %(levelname)s %(message)s")
            handler = logging.StreamHandler()
            handler.setFormatter(formatter)
            logger.addHandler(handler)
            self.new_handlers.append(handler)

        self.logger = logger

    def __enter__(self) -> logging.Logger:
        return self.logger

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        self.logger.setLevel(self.old_level)
        for handler in self.new_handlers:
            self.logger.removeHandler(handler)


legacy_client.LoggingContext = QuietEra5LegacyClientLoggingContext


def load_api_key() -> str:
    """Load the CDS API key from environment variable or from the ~/.cdsapirc file."""
    global _api_key
    if _api_key is not None:
        return _api_key

    if os.getenv("CDSADS_API_KEY"):
        _api_key = os.getenv("CDSADS_API_KEY")
        return _api_key

    try:
        with open(os.path.expanduser("~/.cdsapirc")) as file:
            api_key = file.readlines()[-1].split(":")[-1].strip()
            assert api_key, "CDS API key is empty. Please check your ~/.cdsapirc file."
            _api_key = api_key
            return _api_key
    except FileNotFoundError:
        raise FileNotFoundError(
            "CDS API key file not found. Please create '~/.cdsapirc' with your key."
        )


def now_utc() -> datetime.datetime:
    """Get the current UTC time."""
    return datetime.datetime.now(datetime.UTC)


def make_cds_days_list(year, month) -> list[str]:
    """Generate a list of days for a given year and month.

    In case the month is in the future, return an empty list.

    :param year: The year for which to generate the days.
    :param month: The month for which to generate the days.
    :return: A list of days in the format 'DD' for the specified month and year.
    """

    now = now_utc()
    current_year, current_month, current_day = now.year, now.month, now.day

    if (current_year, current_month) < (year, month):
        return []

    days_in_month = monthrange(year, month)[1]

    if (current_year, current_month) == (year, month):
        return [f"{day:02d}" for day in range(1, min(days_in_month, current_day) + 1)]

    else:
        return [f"{day:02d}" for day in range(1, days_in_month + 1)]


def execute_download_request(url, dataset, cds_request, target_file, verbose: bool = False):
    """Execute a CDS request and download the data to the target file."""
    client = cdsapi.Client(url=url, key=load_api_key(), quiet=(not verbose))
    # wait for a random time between 0 and 10 seconds to avoid hitting the CDS API too hard
    time.sleep(random.uniform(0, 10))
    # Execute the CDS request
    logging.debug(f"Executing CDS request for dataset '{dataset}' with parameters: {cds_request}")
    client.retrieve(dataset, cds_request).download(target=target_file)


def load_netcdf(file_path) -> xr.Dataset:
    """Load a NetCDF file and return its content.

    :param file_path: Path to the NetCDF file.
    :return: xarray.Dataset containing the data from the NetCDF file.
    """
    ds = xr.open_dataset(file_path, engine="netcdf4")  # type: ignore[call-arg]

    return ds


def unzip_and_load_netcdf_to_df(file_path: str, clean_up: bool) -> pd.DataFrame:
    """Unzip a zip file containing a NetCDF file and load it into a DataFrame.

    :param file_path: Path to the zip file.
    :param clean_up: If True, remove the temporary NetCDF file after processing.
    :return: A DataFrame containing the data from the NetCDF file.
    """

    if zipfile.is_zipfile(file_path):
        with zipfile.ZipFile(file_path, "r") as zip_ref:
            nc_file = zip_ref.namelist()[0]
            zip_ref.extractall("/tmp")

        tmp_file_path = os.path.join("/tmp", nc_file)
        try:
            return concat_netcdf_files_to_df(tmp_file_path, time_dim=0)
        finally:
            if clean_up:
                os.remove(tmp_file_path)

    else:
        # If the file is not a zip, assume it's a NetCDF file
        return concat_netcdf_files_to_df(file_path, time_dim=0)


def concat_netcdf_files_to_df(file_paths, time_dim: int = 0) -> pd.DataFrame:
    """Concatenate multiple NetCDF files into a single DataFrame.

    The index of the DataFrame will be a datetime index based on the 'time' dimension

    :param file_paths: List of paths to NetCDF files to merge.
    :param time_dim: The dimension to use for the time index. Default is 0 (first
        dimension).
    :return: xarray.Dataset containing the merged data.
    """
    files = glob.glob(file_paths)
    assert files, f"No NetCDF files found matching pattern: {file_paths}"
    datasets = []
    for file in files:
        try:
            df = xr.open_dataset(file).to_dataframe()
        except Exception as e:
            raise ValueError(f"Failed to read file {file}: {e}")
        # convert index to datetime by keeping the first level (time)
        df.index = pd.to_datetime(df.index.get_level_values(time_dim))
        datasets.append(df)

    combined_ds = pd.concat(datasets, axis=0).sort_index()
    return combined_ds


def generate_download_link(file_path: str, link_text: str = "Download file"):
    """Generate a download link for a file. Useful for Jupyter notebooks to allow users to download
    files generated in the notebook.

    :param file_path: Path to the file to be downloaded.
    :param link_text: Text to be displayed for the download link.
    :return: HTML string containing the download link.
    """
    from IPython.core.display import HTML  # type: ignore

    file_data = Path(file_path).read_bytes()
    return HTML(
        f'<a href="data:application/octet-stream;base64,{b64encode(file_data).decode()}" '
        f'download="{Path(file_path).name}">{link_text}</a>'
    )



import logging
import os.path
from argparse import ArgumentParser
from datetime import datetime

import numpy as np
import pandas as pd
from tqdm.auto import tqdm

from era5epw.ads import download_cams_solar_radiation_data
from era5epw.cds import download_era5_data


def get_first_weekday_of_year(y: int) -> str:
    first_weekday_of_year = pd.Timestamp(y, 1, 1).dayofweek
    return ["Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday", "Sunday"][
        first_weekday_of_year
    ]


def calc_rh(dry_bulb_temp: np.ndarray, dew_point_temp: np.ndarray) -> np.ndarray:
    """Calculate relative humidity from temperature and dew point temperature.

    :param dry_bulb_temp: Temperature in Celsius.
    :param dew_point_temp: Dew point temperature in Celsius.
    """
    es = 6.112 * np.exp((17.67 * dry_bulb_temp) / (dry_bulb_temp + 243.5))
    esd = 6.112 * np.exp((17.67 * dew_point_temp) / (dew_point_temp + 243.5))
    return np.round(100 * esd / es, 1)


def calc_monthly_soil_temperature(soil_temp: pd.DataFrame | pd.Series) -> pd.Series:
    """Compute monthly average soil temperature at 0-7 cm depth (level 1)

    :param soil_temp: Soil temperature data at 0-7 cm depth in Kelvin.
    :return: Monthly average soil temperature in Celsius.
    """
    return soil_temp.groupby(soil_temp.index.month).mean() - 273.15


def is_leap_year(y: int) -> bool:
    """Check if a year is a leap year."""
    return (y % 4 == 0 and y % 100 != 0) or (y % 400 == 0)


def make_data_period_end_date(df: pd.DataFrame) -> str:
    """Return the data period end date string for the EPW header.

    :param df: DataFrame containing the weather data.
    :return: End date string in the format "MM/DD".
    """
    return df.iloc[-1][["Month", "Day"]].astype(int).astype(str).str.cat(sep="/")


def create_args() -> ArgumentParser:
    """Create argument parser for command line arguments."""
    import argparse

    parser = argparse.ArgumentParser(
        description="Generate a full year EPW file from ERA5 and CAMS data."
    )
    parser.add_argument(
        "--year", type=int, default=2024, help="Year for which to generate the EPW file."
    )
    parser.add_argument("--latitude", type=float, default=49.5, help="Latitude of the location.")
    parser.add_argument("--longitude", type=float, default=2.5, help="Longitude of the location.")
    parser.add_argument(
        "--parallel-requests",
        type=int,
        default=10,
        required=False,
        help="Number of parallel requests to make on CDS API.",
    )
    parser.add_argument(
        "--city-name", type=str, default="Paris", help="Name of the city for the EPW file."
    )
    parser.add_argument("--time-zone", type=int, default=1, help="Time zone offset from UTC.")
    parser.add_argument(
        "--elevation", type=int, default=35, help="Elevation of the location in meters."
    )
    parser.add_argument(
        "--output_file",
        type=str,
        default=f"/tmp/era5epw_{datetime.now().strftime('%Y%m%d_%H%M%S')}.epw",
        help="Output file path for the generated EPW file.",
    )
    parser.add_argument(
        "--verbose",
        action="store_true",
        help="Enable verbose logging from CDS client.",
    )
    parser.add_argument(
        "--apply-time-zone-to-data",
        action="store_true",
        help="Apply time zone offset to data timestamps. If false (default), UTC time is kept.",
    )
    return parser


def download_and_make_epw(
    year: int,
    latitude: float,
    longitude: float,
    city_name: str,
    time_zone: int,
    elevation: int,
    output_file: str,
    parallel_exec_nb: int = 10,
    verbose: bool = False,
    apply_time_zone_to_data: bool = False,
) -> None:
    """Generate a full year EPW file from ERA5 and CAMS data.

    :param year: Year for which to generate the EPW file.
    :param latitude: Latitude of the location.
    :param longitude: Longitude of the location.
    :param city_name: Name of the city for the EPW file.
    :param time_zone: Time zone offset from UTC.
    :param elevation: Elevation of the location in meters.
    :param output_file: Path to save the generated EPW file.
    :param parallel_exec_nb: Number of parallel requests to make on CDS (ERA5 allows 10 per
        minute).
    :param verbose: If True, enable verbose logging from CDS client.
    :param apply_time_zone_to_data: If True, apply time zone offset to data timestamps.
    """
    start_time = datetime.now()

    # Log whether time zone is being applied to data
    if apply_time_zone_to_data:
        logging.info(
            f"Time zone offset of {time_zone:+d} hours will be applied to data timestamps."
        )
    else:
        logging.info(
            f"Data will use UTC time. Time zone offset of {time_zone:+d} hours "
            "will only be used in EPW LOCATION header."
        )

    # Create overall progress bar for the two main download phases
    overall_progress = tqdm(total=2, desc="Overall progress", unit="phase", position=0)

    overall_progress.set_description("Downloading CAMS solar radiation data")
    cams_df = download_cams_solar_radiation_data(
        longitude=longitude,
        latitude=latitude,
        year=year,
        time_zone=time_zone if apply_time_zone_to_data else None,
    )
    overall_progress.update(1)  # CAMS download completed

    overall_progress.set_description("Downloading ERA5 data")
    era5_df = download_era5_data(
        variables=[
            "2m_temperature",
            "2m_dewpoint_temperature",
            "surface_pressure",
            "10m_u_component_of_wind",
            "10m_v_component_of_wind",
            "total_cloud_cover",
            "uv_visible_albedo_for_direct_radiation",
            "snow_depth",
            "soil_temperature_level_1",
            "total_precipitation",
        ],
        year=year,
        latitude=latitude,
        longitude=longitude,
        parallel_exec_nb=parallel_exec_nb,
        dataset=None,  # dynamic dataset selection based on variables
        verbose=verbose,
        time_zone=time_zone if apply_time_zone_to_data else None,
    )
    overall_progress.update(1)  # ERA5 download completed
    overall_progress.close()

    # Apply time zone shift if requested
    if apply_time_zone_to_data:
        logging.info(f"Applying time zone offset of {time_zone:+d} hours to data timestamps.")
        era5_df.index = era5_df.index + pd.Timedelta(hours=time_zone)
        cams_df.index = cams_df.index + pd.Timedelta(hours=time_zone)

    # Filter to keep only data within the target year
    # (filters out extra days added for time zone or to accommodate with missing first hour)
    series_start = f"{year}-01-01 00:00:00"
    series_end = f"{year}-12-31 23:00:00"
    era5_df = era5_df.truncate(before=series_start, after=series_end)
    cams_df = cams_df.truncate(before=series_start, after=series_end)

    # Align ERA5 and CAMS dataframes to the same time range
    # their indices may not match exactly, especially when the year is not complete (e.g. current year)
    era5_df, cams_df = era5_df.align(cams_df, join="inner", axis=0)
    assert era5_df.index.equals(cams_df.index), "Time indices of ERA5 and CAMS data do not match"

    # Extract variables, convert to correct units
    temp_C = era5_df["t2m"].values - 273.15  # K to C
    dew_C = era5_df["d2m"].values - 273.15  # K to C
    press = era5_df["sp"].values  # Pa
    u10 = era5_df["u10"].values  # m/s
    v10 = era5_df["v10"].values  # m/s
    cloud = era5_df["tcc"].values * 10  # Fraction to okta (0-10 scale)
    uv_visible_albedo = era5_df["aluvp"].values  # (0-1 scale)
    snow_depth = era5_df["sd"].values * 100  # m to cm
    total_precipitation = era5_df["tp"].values * 1000  # m to mm
    ghi = cams_df["GHI"].values  # Global horizontal all sky irradiation in Wh/m^2
    bni = cams_df["BNI"].values  # Direct normal all sky irradiation in Wh/m^2
    bhi = cams_df["BHI"].values  # Direct horizontal all sky irradiation in Wh/m^2
    dhi = cams_df["DHI"].values  # Diffuse horizontal irradiation in Wh/m^2

    # Calculate wind speed and direction
    wind_speed = np.sqrt(u10**2 + v10**2)
    wind_dir = (180 + np.degrees(np.arctan2(u10, v10))) % 360

    # Time index
    times = pd.to_datetime(era5_df.index.values)

    # Create DataFrame
    df = pd.DataFrame(
        {
            "Year": times.year,
            "Month": times.month,
            "Day": times.day,
            "Hour": times.hour + 1,  # EPW hours start at 1
            "Minute": 0,
            "Data Source and Uncertainty Flags": "9",
            "Dry Bulb Temperature": np.round(temp_C, 1),  # C
            "Dew Point Temperature": np.round(dew_C, 1),  # C
            "Relative Humidity": calc_rh(temp_C, dew_C),  # %
            "Atmospheric Station Pressure": np.round(press, 0),  # Pa
            "Extraterrestrial Horizontal Radiation": 9999,  # Wh/m^2
            "Extraterrestrial Direct Normal Radiation": 9999,  # Wh/m^2
            "Horizontal Infrared Radiation Intensity": 9999,  # Wh/m^2 - TODO
            "Global Horizontal Radiation": np.round(ghi, 1),  # Wh/m^2
            "Direct Normal Radiation": np.round(bni, 1),  # Wh/m^2
            "Diffuse Horizontal Radiation": np.round(dhi, 1),  # Wh/m^2
            "Global Horizontal Illuminance": np.round(110 * ghi, 0),  # Lux
            "Direct Normal Illuminance": np.round(105 * bni, 0),  # Lux
            "Diffuse Horizontal Illuminance": np.round(119 * dhi, 0),  # Lux
            "Zenith Luminance": 9999,  # Cd/m^2 - TODO
            "Wind Direction": np.round(wind_dir, 0),  # degrees
            "Wind Speed": np.round(wind_speed, 1),  # m/s
            "Total Sky Cover": np.round(cloud, 0),
            "Opaque Sky Cover": np.round(cloud, 0),
            "Visibility": 9999,  # km
            "Ceiling Height": 77777,  # m - TODO
            # 0 = Weather observation made; 9 = Weather observation not made, or missing
            "Present Weather Observation": 0,
            "Present Weather Codes": 999999999,  # see doc
            "Precipitable Water": 999,  # mm
            "Aerosol Optical Depth": 999,  # thousandths
            "Snow Depth": np.round(snow_depth, 1),  # cm
            "Days Since Last Snowfall": 99,
            "Albedo": np.round(uv_visible_albedo, 1),  # (0 - 1 scale)
            "Liquid Precipitation Depth": np.round(total_precipitation, 1),  # mm
            "Liquid Precipitation Quantity": 1,
        }
    )

    ground_temps = "1,3.5,,,," + ",".join(
        calc_monthly_soil_temperature(era5_df["stl1"]).round(1).astype(str).tolist()
    )
    if "nan" in ground_temps:
        logging.warning(
            "Soil temperature data at level 1 (0-7 cm) contains NaN values. "
            "Setting number of monthly soil temperatures to 0."
        )
        ground_temps = "0"

    data_period_end_date = make_data_period_end_date(df)

    # Write header
    epw_header = [
        f"LOCATION,{city_name},,,ERA5 (ECMWF),n/a,{latitude:.2f},{longitude:.2f},{time_zone},{elevation}",
        "DESIGN CONDITIONS,0",
        "TYPICAL/EXTREME PERIODS,0",
        f"GROUND TEMPERATURES,{ground_temps}",
        f"HOLIDAYS/DAYLIGHT SAVINGS,{'Yes' if is_leap_year(year) else 'No'},0,0,0",
        "COMMENTS 1,Data from ERA5 and CAMS via CDSAPI",
        "COMMENTS 2,Processed with Python - Provided with love by the Foobot Team",
        f"DATA PERIODS,1,1,Data,{get_first_weekday_of_year(year)},1/1,{data_period_end_date}",
    ]

    with open(output_file, "w") as f:
        for line in epw_header:
            f.write(line + os.linesep)
        df.to_csv(f, index=False, header=False, lineterminator=os.linesep)

    end_time = datetime.now()
    tqdm.write(f"EPW file written as {output_file}. Took {end_time - start_time} to generate.")


def download():
    from era5epw.logcfg import init_logging

    args = create_args().parse_args()

    # Initialize logging with verbosity setting
    init_logging(verbose=args.verbose)

    tqdm.write(
        f"Generating EPW file for {args.city_name} ({args.latitude}, {args.longitude}) in {args.year}..."
    )

    if os.path.exists(args.output_file):
        tqdm.write(f"Output file {args.output_file} already exists. It will be overwritten.")

    download_and_make_epw(
        year=args.year,
        latitude=args.latitude,
        longitude=args.longitude,
        city_name=args.city_name,
        time_zone=args.time_zone,
        elevation=args.elevation,
        output_file=args.output_file,
        parallel_exec_nb=args.parallel_requests,
        verbose=args.verbose,
        apply_time_zone_to_data=args.apply_time_zone_to_data,
    )


if __name__ == "__main__":
    download_and_make_epw(
        year=2024,
        latitude=49.4,
        longitude=0.1,
        city_name="Le Havre",
        time_zone=1,
        elevation=0,
        output_file=f"/tmp/era5epw_{datetime.now().strftime('%Y%m%d_%H%M%S')}.epw",
        parallel_exec_nb=10,
        verbose=False,
        apply_time_zone_to_data=False,
    )



import logging.config


def init_logging(verbose: bool = False) -> None:
    # Set CDS client logger levels based on verbosity
    cds_log_level = "INFO" if verbose else "ERROR"
    app_log_level = "INFO" if verbose else "WARNING"

    logging.config.dictConfig(
        {
            "version": 1,
            "disable_existing_loggers": True,
            "formatters": {
                "standard": {
                    "format": (
                        "%(asctime)s "
                        + "[%(levelname)s]"
                        + "[%(filename)s:%(lineno)s - %(funcName)s()]: %(message)s"
                    )
                },
            },
            "handlers": {
                "default": {
                    "level": "INFO",
                    "formatter": "standard",
                    "class": "logging.StreamHandler",
                    "stream": "ext://sys.stdout",  # Default is stderr
                },
            },
            "loggers": {
                "": {"handlers": ["default"], "level": app_log_level},  # root logger
                "cdsapi": {"handlers": ["default"], "level": cds_log_level, "propagate": False},
                "ecmwf.datastores.legacy_client": {
                    "handlers": ["default"],
                    "level": cds_log_level,
                    "propagate": False,
                },
                "multiurl": {"handlers": ["default"], "level": cds_log_level, "propagate": False},
            },
        }
    )



"""EPW file visualization tool.

This module provides interactive visualization of EPW (EnergyPlus Weather) files using
Plotly. It supports both command-line usage and API usage in Jupyter notebooks.
"""

import argparse
from pathlib import Path
from typing import Literal

import pandas as pd
import plotly.graph_objects as go

# EPW column names as per EnergyPlus EPW format specification
EPW_COLUMNS = [
    "Year",
    "Month",
    "Day",
    "Hour",
    "Minute",
    "Data Source and Uncertainty Flags",
    "Dry Bulb Temperature",
    "Dew Point Temperature",
    "Relative Humidity",
    "Atmospheric Station Pressure",
    "Extraterrestrial Horizontal Radiation",
    "Extraterrestrial Direct Normal Radiation",
    "Horizontal Infrared Radiation Intensity",
    "Global Horizontal Radiation",
    "Direct Normal Radiation",
    "Diffuse Horizontal Radiation",
    "Global Horizontal Illuminance",
    "Direct Normal Illuminance",
    "Diffuse Horizontal Illuminance",
    "Zenith Luminance",
    "Wind Direction",
    "Wind Speed",
    "Total Sky Cover",
    "Opaque Sky Cover",
    "Visibility",
    "Ceiling Height",
    "Present Weather Observation",
    "Present Weather Codes",
    "Precipitable Water",
    "Aerosol Optical Depth",
    "Snow Depth",
    "Days Since Last Snowfall",
    "Albedo",
    "Liquid Precipitation Depth",
    "Liquid Precipitation Quantity",
]

# Series to hide from the dropdown menu
HIDDEN_SERIES = [
    "Year",
    "Month",
    "Day",
    "Hour",
    "Minute",
    "Data Source and Uncertainty Flags",
    "Present Weather Observation",
    "Present Weather Codes",
]

# Units for each series
UNITS = {
    "Dry Bulb Temperature": "°C",
    "Dew Point Temperature": "°C",
    "Relative Humidity": "%",
    "Atmospheric Station Pressure": "Pa",
    "Extraterrestrial Horizontal Radiation": "Wh/m²",
    "Extraterrestrial Direct Normal Radiation": "Wh/m²",
    "Horizontal Infrared Radiation Intensity": "Wh/m²",
    "Global Horizontal Radiation": "Wh/m²",
    "Direct Normal Radiation": "Wh/m²",
    "Diffuse Horizontal Radiation": "Wh/m²",
    "Global Horizontal Illuminance": "lux",
    "Direct Normal Illuminance": "lux",
    "Diffuse Horizontal Illuminance": "lux",
    "Zenith Luminance": "Cd/m²",
    "Wind Direction": "degrees",
    "Wind Speed": "m/s",
    "Total Sky Cover": "",
    "Opaque Sky Cover": "",
    "Visibility": "km",
    "Ceiling Height": "m",
    "Precipitable Water": "mm",
    "Aerosol Optical Depth": "thousandths",
    "Snow Depth": "cm",
    "Days Since Last Snowfall": "",
    "Albedo": "",
    "Liquid Precipitation Depth": "mm",
    "Liquid Precipitation Quantity": "hr",
}


def read_epw_file(epw_file_path: str) -> pd.DataFrame:
    """Read and parse an EPW file into a DataFrame.

    :param epw_file_path: Path to the EPW file.
    :return: DataFrame with datetime index and weather data columns.
    """
    with open(epw_file_path) as f:
        # Skip the 8 header lines
        lines = f.readlines()

    # Data starts from line 9 (index 8)
    data_lines = lines[8:]

    # Parse data lines
    data = []
    for line in data_lines:
        if line.strip():
            data.append(line.strip().split(","))

    # Create DataFrame
    df = pd.DataFrame(data, columns=EPW_COLUMNS)

    # Convert numeric columns to float
    for col in df.columns:
        if col not in ["Year", "Month", "Day", "Hour", "Minute"]:
            df[col] = pd.to_numeric(df[col], errors="coerce")

    # Convert time columns to int
    df["Year"] = df["Year"].astype(int)
    df["Month"] = df["Month"].astype(int)
    df["Day"] = df["Day"].astype(int)
    df["Hour"] = df["Hour"].astype(int)

    # Create datetime index (handle hour 24 as hour 0 of next day)
    def create_datetime(row):
        year, month, day, hour = row["Year"], row["Month"], row["Day"], row["Hour"]
        if hour == 24:
            # Hour 24 means midnight of the next day
            dt = pd.Timestamp(year=year, month=month, day=day) + pd.Timedelta(days=1)
        else:
            dt = pd.Timestamp(year=year, month=month, day=day, hour=hour)
        return dt

    df["Datetime"] = df.apply(create_datetime, axis=1)
    df.set_index("Datetime", inplace=True)

    return df


def get_visible_series(df: pd.DataFrame) -> list[str]:
    """Get list of visible series names (excluding hidden ones).

    :param df: DataFrame containing weather data.
    :return: List of visible series names.
    """
    return [col for col in df.columns if col not in HIDDEN_SERIES]


def create_2d_plot(
    df: pd.DataFrame, series_name: str, width: int = 1200, height: int = 500
) -> go.Figure:
    """Create a 2D line plot for a given weather series.

    :param df: DataFrame containing weather data.
    :param series_name: Name of the series to plot.
    :param width: Width of the plot in pixels.
    :param height: Height of the plot in pixels.
    :return: Plotly Figure object.
    """
    unit = UNITS.get(series_name, "")

    fig = go.Figure()
    fig.add_trace(
        go.Scattergl(
            x=df.index,
            y=df[series_name],
            mode="lines",
            name=series_name,
            line=dict(color="rgb(31, 119, 180)"),
        )
    )

    fig.update_layout(
        title=dict(text=f"{series_name} Over Time", font=dict(size=16)),
        xaxis=dict(title="Date", automargin=True),
        yaxis=dict(title=f"{series_name} ({unit})" if unit else series_name, automargin=True),
        width=width,
        height=height,
        margin=dict(t=50, r=20, b=50, l=60),
        hovermode="x unified",
    )

    return fig


def create_3d_plot(
    df: pd.DataFrame, series_name: str, width: int = 1200, height: int = 750
) -> go.Figure:
    """Create a 3D surface plot showing hour vs day with data values.

    :param df: DataFrame containing weather data.
    :param series_name: Name of the series to plot.
    :param width: Width of the plot in pixels.
    :param height: Height of the plot in pixels.
    :return: Plotly Figure object.
    """
    unit = UNITS.get(series_name, "")

    # Reshape data into 2D array: rows = days, columns = hours
    per_chunk = 24
    values = df[series_name].values

    # Chunk data into days (24 hours per day)
    z_data = []
    for i in range(0, len(values), per_chunk):
        chunk = values[i : i + per_chunk]
        if len(chunk) == per_chunk:
            z_data.append(chunk)

    # Get the first date
    day_0 = df.index[0].date()

    # Create y-axis: dates for each day
    y_dates = [day_0 + pd.Timedelta(days=i) for i in range(len(z_data))]

    fig = go.Figure(
        data=[
            go.Surface(
                y=y_dates,
                x=list(range(24)),
                z=z_data,
                colorscale="Jet",
                colorbar=dict(len=0.5, title=unit if unit else ""),
            )
        ]
    )

    fig.update_layout(
        scene=dict(
            camera=dict(
                eye=dict(x=1.87, y=0.88, z=0.64),
                up=dict(x=0, y=0, z=1),
                center=dict(x=0, y=0, z=-0.35),
                projection=dict(type="perspective"),
            ),
            xaxis=dict(
                title="Hour",
                ticksuffix="h",
                autorange="reversed",
            ),
            yaxis=dict(
                title="Day of Year",
            ),
            zaxis=dict(
                title=f"{series_name} ({unit})" if unit else series_name,
            ),
            aspectratio=dict(x=1, y=2, z=0.5),
        ),
        title=dict(text=f"{series_name} - 3D View", font=dict(size=16)),
        autosize=False,
        width=width,
        height=height,
        margin=dict(l=20, r=50, b=5, t=50),
    )

    return fig


def create_radar_plot(
    df: pd.DataFrame, series_name: str, width: int = 900, height: int = 700
) -> go.Figure:
    """Create a radar (polar) plot showing daily min/max values throughout the year.

    :param df: DataFrame containing weather data.
    :param series_name: Name of the series to plot.
    :param width: Width of the plot in pixels.
    :param height: Height of the plot in pixels.
    :return: Plotly Figure object.
    """
    unit = UNITS.get(series_name, "")

    # Group by date to get daily min/max
    daily_data = df.groupby(df.index.date)[series_name].agg(["min", "max"])
    daily_data["date"] = daily_data.index
    daily_data["month"] = pd.to_datetime(daily_data["date"]).dt.strftime("%b")

    # Get max value for color scaling
    max_val = daily_data["max"].max()
    min_val = daily_data["min"].min()
    val_range = max_val - min_val if max_val != min_val else 1

    # Jet colorscale function
    def jet_colorscale(normalized_value: float, alpha: float = 1.0) -> str:
        """Create a jet colorscale color."""
        val = max(0, min(1, normalized_value))

        if val < 0.25:
            r, g, b = 0, int(4 * val * 255), 255
        elif val < 0.5:
            r, g, b = 0, 255, int((1 + 4 * (0.25 - val)) * 255)
        elif val < 0.75:
            r, g, b = int(4 * (val - 0.5) * 255), 255, 0
        else:
            r, g, b = 255, int((1 + 4 * (0.75 - val)) * 255), 0

        r = max(0, min(255, r))
        g = max(0, min(255, g))
        b = max(0, min(255, b))

        return f"rgba({r}, {g}, {b}, {alpha})"

    # Create traces for each day
    traces = []
    max_angle = min(360, len(daily_data))

    # Track month changes for labels
    month_ticks = []
    current_month = ""

    for i, (idx, row) in enumerate(daily_data.iterrows()):
        theta = (i / len(daily_data)) * max_angle

        # Color based on max temperature
        color_val = (row["max"] - min_val) / val_range if val_range > 0 else 0
        color = jet_colorscale(color_val, 1.0)

        # Add month label only when month changes
        month_str = row["month"]
        if month_str != current_month:
            month_ticks.append((i, f"<b>{month_str}</b>"))
            current_month = month_str
        else:
            month_ticks.append((i, ""))

        traces.append(
            go.Scatterpolargl(
                r=[row["max"], row["min"]],
                theta=[theta, theta],
                mode="lines",
                line=dict(color=color, width=3),
                fill="toself",
                name=str(row["date"]),
                hovertemplate=(
                    f"{row['date']}<br>"
                    f"Min: {row['min']:.1f}{unit}<br>"
                    f"Max: {row['max']:.1f}{unit}<extra></extra>"
                ),
                showlegend=False,
            )
        )

    fig = go.Figure(data=traces)

    fig.update_layout(
        polar=dict(
            radialaxis=dict(
                visible=True,
                layer="above traces",
                title=dict(
                    text=f"{series_name} ({unit})" if unit else series_name,
                    font=dict(size=15),
                ),
                tickfont=dict(size=14),
                gridcolor="lightgray",
                linecolor="gray",
                linewidth=4,
            ),
            angularaxis=dict(
                direction="clockwise",
                gridcolor="lightgray",
                linecolor="gray",
                rotation=90,  # Start first day at top
                tickmode="array",
                tickvals=[t[0] for t in month_ticks],
                ticktext=[t[1] for t in month_ticks],
                tickangle=0,
                ticks="inside",
            ),
        ),
        showlegend=False,
        margin=dict(t=80, r=50, b=50, l=50),
        width=width,
        height=height,
        title=dict(
            text=f"Daily Min/Max {series_name}",
            font=dict(size=16),
            y=0.98,
        ),
        hoverlabel=dict(
            bgcolor="white",
            font=dict(size=14),
        ),
    )

    return fig


def visualize_epw(
    epw_file_path: str,
    series_name: str = "Dry Bulb Temperature",
    plot_type: Literal["2D", "3D", "radar"] = "2D",
    show: bool = True,
    renderer: Literal["notebook", "browser", "iframe"] = "notebook",
) -> go.Figure:
    """Visualize EPW weather data with interactive plots.

    :param epw_file_path: Path to the EPW file.
    :param series_name: Name of the weather series to visualize.
    :param plot_type: Type of plot: "2D", "3D", or "radar".
    :param show: If True, display the plot immediately (useful in Jupyter).
    :param renderer: Renderer to use for displaying the plot ("notebook", "browser", or
        "iframe").
    :return: Plotly Figure object.
    """
    # Read EPW file
    df = read_epw_file(epw_file_path)

    # Validate series name
    visible_series = get_visible_series(df)
    if series_name not in visible_series:
        raise ValueError(
            f"Series '{series_name}' not found. Available series: {', '.join(visible_series)}"
        )

    # Create the appropriate plot
    if plot_type == "2D":
        fig = create_2d_plot(df, series_name)
    elif plot_type == "3D":
        fig = create_3d_plot(df, series_name)
    elif plot_type == "radar":
        fig = create_radar_plot(df, series_name)
    else:
        raise ValueError(f"Invalid plot_type: {plot_type}. Must be '2D', '3D', or 'radar'.")

    if show:
        fig.show(renderer=renderer)

    return fig


def visualize_cli() -> None:
    """Command-line interface for EPW visualization."""
    parser = argparse.ArgumentParser(
        description="Visualize EPW (EnergyPlus Weather) file data with interactive plots."
    )
    parser.add_argument(
        "epw_file",
        type=str,
        help="Path to the EPW file to visualize.",
    )
    parser.add_argument(
        "--series",
        type=str,
        default="Dry Bulb Temperature",
        help="Weather series to visualize (e.g., 'Dry Bulb Temperature', 'Wind Speed').",
    )
    parser.add_argument(
        "--type",
        type=str,
        choices=["2D", "3D", "radar"],
        default="2D",
        help="Type of visualization: 2D (line plot), 3D (surface plot), or radar (polar plot).",
    )
    parser.add_argument(
        "--output",
        type=str,
        help="Path to save the plot as an HTML file. If not provided, opens in browser.",
    )
    parser.add_argument(
        "--list-series",
        action="store_true",
        help="List all available weather series in the EPW file and exit.",
    )

    args = parser.parse_args()

    # Check if file exists
    if not Path(args.epw_file).exists():
        print(f"Error: EPW file not found: {args.epw_file}")
        return

    # List series if requested
    if args.list_series:
        df = read_epw_file(args.epw_file)
        visible_series = get_visible_series(df)
        print("Available weather series:")
        for series in visible_series:
            unit = UNITS.get(series, "")
            print(f"  - {series}" + (f" ({unit})" if unit else ""))
        return

    # Create single visualization
    try:
        fig = visualize_epw(
            epw_file_path=args.epw_file,
            series_name=args.series,
            plot_type=args.type,
            show=False,
        )

        if args.output:
            fig.write_html(args.output)
            print(f"Visualization saved to {args.output}")
        else:
            fig.show()

    except Exception as e:
        print(f"Error creating visualization: {e}")
        import traceback

        traceback.print_exc()


if __name__ == "__main__":
    visualize_cli()
